unit Breakpoint;

(***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************)

interface

uses
  Mars;

const
  ANY_VALUE = High(TOperand);

type
  TBreakpointTriggerType = (BRK_PC,BRK_WRITE);
  TBreakpointTriggerField = (BRK_NA,BRK_AFIELD,BRK_BFIELD,BRK_BOTHFIELDS);
  TBreakpointManager = class; // forward
  TBreakpoint = class
    protected
      FMgr: TBreakpointManager;
      FWarrior: TWarrior;
      FTrigger: TBreakpointTriggerType;
      FTriggered: boolean;
      FPC,
      FAddr,
      FValueA,
      FValueB: TOperand;
      FField: TBreakpointTriggerField;
      constructor Create(AMgr: TBreakpointManager);
      function TriggeredPC(const AThread: TThread): boolean;
      function TriggeredWrite(const AWarrior: TWarrior; const ACore: TInstruction): boolean;
    public
      property Warrior: TWarrior read FWarrior;
      property Trigger: TBreakpointTriggerType read FTrigger;
      property Addr: TOperand read FAddr;
      property PC: TOperand read FPC;
      property Field: TBreakpointTriggerField read FField;
      property ValueA: TOperand read FValueA;
      property ValueB: TOperand read FValueB;
      property Manager: TBreakpointManager read FMgr;
      property Triggered: boolean read FTriggered;
    end;
  TBreakpointListener = procedure (Sender: TBreakpointManager) of object;
  TBreakpointManager = class
    protected
      FMars: TMars;
      FCount: integer;
      FInTrigger: boolean;
      FCurrentWarrior: TWarrior;
      FBreakpoints: array of TBreakpoint;
      FListeners: array of TBreakpointListener;
      function OnCoreChange(const ACore: TInstruction): boolean;
      procedure OnStep(const AThread: Mars.TThread);
      function GetBreakpoint(AIndex: integer): TBreakpoint;
      procedure NotifyListeners;
    public
      constructor Create(AMars: TMars);
      destructor Destroy; override;
      property Mars: TMars read FMars;
      procedure AddBreakpoint(APC: TOperand; AWarrior: TWarrior); overload;
      procedure AddBreakpoint(AAddr: TOperand; AWarrior: TWarrior; AValueA, AValueB: TOperand); overload;
      procedure RemoveBreakpoint(ABreakpoint: TBreakpoint);
      property Breakpoint[AIndex: integer]: TBreakpoint read GetBreakpoint;
      property Count: integer read FCount;
      procedure AddBreakpointListener(AListener: TBreakpointListener);
      procedure RemoveBreakpointListener(AListener: TBreakpointListener);
    end;

implementation

(****** TBreakpoint implementation *********************************)

constructor TBreakpoint.Create(AMgr: TBreakpointManager);
begin
  FMgr := AMgr;
  FWarrior := nil;
  FTrigger := BRK_PC;
  FPC := ANY_VALUE;
  FAddr := ANY_VALUE;
  FField := BRK_NA;
  FValueA := ANY_VALUE;
  FValueB := ANY_VALUE;
end;

function TBreakpoint.TriggeredPC(const AThread: TThread): boolean;
begin
  Result := ((FTrigger = BRK_PC) and (AThread.PC = FPC));
  if Assigned(FWarrior) and (AThread.Process.Warrior <> FWarrior) then
    Result := False;
end;

function TBreakpoint.TriggeredWrite(const AWarrior: TWarrior; const ACore: TInstruction): boolean;
begin
  Result := False;
  if (FTrigger <> BRK_WRITE) then
    Exit;
  if Assigned(FWarrior) and (AWarrior <> FWarrior) then
    Exit;
  if (FAddr <> ACore.Index) then
    Exit;
  if (FField in [BRK_AFIELD,BRK_BOTHFIELDS]) and (ACore.A.Value <> FValueA) and (FValueA <> ANY_VALUE) then
    Exit;
  if (FField in [BRK_BFIELD,BRK_BOTHFIELDS]) and (ACore.B.Value <> FValueB) and (FValueB <> ANY_VALUE) then
    Exit;
  Result := True;
end;

(*** TBreakpointManager implementation *********************************)

constructor TBreakpointManager.Create(AMars: TMars);
begin
  FMars := AMars;
  FCount := 0;
  SetLength(FBreakpoints,FCount);
  FInTrigger := False;
  FCurrentWarrior := nil;
  // register my listeners
  FMars.AddCoreListener(OnCoreChange);
  FMars.AddStepListener(OnStep);
end;

destructor TBreakpointManager.Destroy;
begin
  FMars.RemoveCoreListener(OnCoreChange);
  FMars.RemoveStepListener(OnStep);
  FMars := nil;
  inherited;
end;

procedure TBreakpointManager.AddBreakpoint(APC: TOperand; AWarrior: TWarrior);
begin
  SetLength(FBreakpoints,Succ(FCount));
  FBreakpoints[FCount] := TBreakpoint.Create(Self);
  with FBreakpoints[FCount] do begin
    FTrigger := BRK_PC;
    FWarrior := AWarrior;
    FPC := APC;
  end;
  Inc(FCount);
  NotifyListeners;
end;

procedure TBreakpointManager.AddBreakpoint(AAddr: TOperand; AWarrior: TWarrior; AValueA, AValueB: TOperand);
begin
  SetLength(FBreakpoints,Succ(FCount));
  FBreakpoints[FCount] := TBreakpoint.Create(Self);
  with FBreakpoints[FCount] do begin
    FTrigger := BRK_WRITE;
    FWarrior := AWarrior;
    FAddr := AAddr;
    FField := BRK_NA;
    FValueA := AValueA;
    FValueB := AValueB;
    if (FValueA <> ANY_VALUE) then
      FField := BRK_AFIELD;
    if (FValueB <> ANY_VALUE) then
      if (FField = BRK_AFIELD) then
        FField := BRK_BOTHFIELDS
      else
        FField := BRK_BFIELD;
  end;
  Inc(FCount);
  NotifyListeners;
end;

procedure TBreakpointManager.RemoveBreakpoint(ABreakpoint: TBreakpoint);
function GetBreakpoint: integer;
begin
  for Result := Low(FBreakpoints) to High(FBreakpoints) do
    if (@FBreakpoints[Result] = @ABreakpoint) then
      Exit;
  Result := -1;
end;
var
  i,j, Len: integer;
begin
  i := GetBreakpoint;
  if (i = -1) then Exit;
  Len := Pred(Length(FBreakpoints));
  for j := Succ(i) to Pred(Len) do
    FBreakpoints[j] := FBreakpoints[Succ(j)];
  SetLength(FBreakpoints,Len);
  ABreakpoint.Free;
end;

function TBreakpointManager.GetBreakpoint(AIndex: integer): TBreakpoint;
begin
  {$RANGECHECKS ON}
  Result := FBreakpoints[AIndex];
  {$RANGECHECKS OFF}
end;

function TBreakpointManager.OnCoreChange(const ACore: TInstruction): boolean;
var
  i: integer;
begin
  Result := False;
  for i := Low(FBreakpoints) to High(FBreakpoints) do
    with FBreakpoints[i] do begin
      FTriggered := (FTriggered or TriggeredWrite(FCurrentWarrior,ACore)); // note the OR here!
      Result := (FTriggered or Result);
    end;
  if Result then begin
    FMars.State := RUN_STEP; // prevent us from running further!
    NotifyListeners;
  end;
  FInTrigger := (FInTrigger or Result);
end;

procedure TBreakpointManager.OnStep(const AThread: Mars.TThread);
var
  i: integer;
  Trig: boolean;
begin
  Trig := False;
  if assigned(AThread) then begin
    FCurrentWarrior := AThread.Process.Warrior;
    for i := Low(FBreakpoints) to High(FBreakpoints) do
      with FBreakpoints[i] do begin
        FTriggered := TriggeredPC(AThread);
        Trig := (Trig or FTriggered);
      end;
    if Trig then begin
      FMars.State := RUN_STEP; // prevent us from running further!
      NotifyListeners;
    end;
  end else
    FCurrentWarrior := nil;
  if FInTrigger and (not Trig) then
    NotifyListeners; // unselect them
  FInTrigger := Trig;
end;

procedure TBreakpointManager.AddBreakpointListener(AListener: TBreakpointListener);
var
  i: integer;
begin
  i := Length(FListeners);
  SetLength(FListeners,Succ(i));
  FListeners[i] := AListener;
end;

procedure TBreakpointManager.RemoveBreakpointListener(AListener: TBreakpointListener);
function GetListener: integer;
begin
  for Result := Low(FListeners) to High(FListeners) do begin
    if (@FListeners[Result] = @AListener) then
      Exit;
  end;
  Result := -1; // not found
end;
var
  i, Len: integer;
begin
  i := GetListener;
  if (i = -1) then // no such listener?
    Exit;
  Len := Pred(Length(FListeners));
  FListeners[i] := FListeners[Len];
  SetLength(FListeners,Len);
end;

procedure TBreakpointManager.NotifyListeners;
var
  i: integer;
begin
  for i := Low(FListeners) to High(FListeners) do
    FListeners[i](Self);
end;

end.
