unit Mars;

(***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************)

interface

uses
  SysUtils;

type
  EMars = class(Exception); // thrown if there is any issue with anything
  TOperand = integer;
  TOpCode = (
    oDAT, // terminate process
    oMOV, // move from A to B
    oADD, // add A to B, store result in B
    oSUB, // subtract A from B, store result in B
    oMUL, // multiply A by B, store result in B
    oDIV, // divide B by A, store result in B if A <> 0, else terminate
    oMOD, // divide B by A, store remainder in B if A <> 0, else terminate
    oJMP, // transfer execution to A
    oJMZ, // transfer execution to A if B is zero
    oJMN, // transfer execution to A if B is non-zero
    oDJN, // decrement B, if B is non-zero, transfer execution to A
    oSPL, // split off process to A
    oSLT, // skip next instruction if A is less than B
    oCMP, // same as SEQ
    oSEQ = oCMP, // (*) Skip next instruction if A is equal to B
    oSNE, // (*) Skip next instruction if A is not equal to B
    oNOP, // (*) No operation
    oLDP, // (+) Load P-space cell A into core address B
    oSTP  // (+) Store A-number into P-space cell B
    );
  TOpModifier = (
    mA, // Instructions read and write A-fields.
    mB, // Instructions read and write B-fields.
    mAB, { Instructions read the A-field of the A-instruction  and
         the B-field of the B-instruction and write to B-fields.}
    mBA, { Instructions read the B-field of the A-instruction  and
         the A-field of the B-instruction and write to A-fields. }
    mF, { Instructions read both A- and B-fields of  the  the  A-
         and  B-instruction and write to both A- and B-fields (A
         to A and B to B).}
    mX, { Instructions read both A- and B-fields of  the  the  A-
         and  B-instruction  and  write  to both A- and B-fields
         exchanging fields (A to B and B to A).}
    mI  // Instructions read and write entire instructions.
    );
  TAddressingMode = (
    aImmediate,  // #       immediate
    aDirect,     // $       direct
    aIndirectB,  // @       indirect using B-field
    aPreDecB,    // <       predecrement indirect using B-field
    aPostIncB,   // >       postincrement indirect using B-field
    aIndirectA,  // *       (*) indirect using A-field
    aPreDecA,    // {       (*) predecrement indirect using A-field
    aPostIncA   // }       (*) postincrement indirect using A-field
    );
  PInstruction = ^TInstruction;
  TWarrior = class; // forward
  TInstruction = record
    Index: TOperand;
    Owner: TWarrior;
    ContainsCode: boolean;
    OpCode: TOpCode;
    OpModifier: TOpModifier;
    A, B: record
      AddrMode: TAddressingMode;
      Value: TOperand;
      end;
    ReadWatchCount: integer;
    end;
  TWarrior = class // this is not in the core; it would be represented in the core by a TProcess instance
    protected
      FID: cardinal;
      FFilename,
      FAuthor,
      FName,
      FHill: string;
      FPin,
      FOrg: TOperand; // origin
      FCore: array of TInstruction;
      function GetCore(AIndex: TOperand): TInstruction;
      function GetName: string;
    public
      constructor Create(AID: cardinal; AFilename: string);
      // these are properties to give me the option of changing style later, but for now they are all read/writable
      property Filename: string read FFilename write FFilename;
      property Author: string read FAuthor write FAuthor;
      property Name: string read GetName write FName;
      property Hill: string read FHill write FHill;
      property ID: cardinal read FID write FID;
      property Org: TOperand read FOrg write FOrg;
      // getters
      property Core[AIndex: TOperand]: TInstruction read GetCore;
      function Length: TOperand;
      property Pin: TOperand read FPin write FPin;
      // adders
      procedure Append(ACore: TInstruction);
    end;
  TWarriors = array of TWarrior;
  TThread = class; // forward
  TProcess = class
    protected
      FAlive: boolean;
      FWarrior: TWarrior;
      FActiveThread, // the index in FThreads of the currently active thread
      FMaxThreads,
      FThreadCount: integer; // number of slots in thread array that are used
      FThreads: array of TThread;
      constructor Create(AWarrior: TWarrior; AMaxThreads: integer);
      procedure Next; // move the thread pointer to the next thread
      function GetActiveThread: TThread; // get the current active thread
      procedure CreateThread(APC: TOperand);
      procedure KillThread; // kill the active thread
    public
      // tags for graphical stuff
      // real stuff
      property Alive: boolean read FAlive;
      function Dead: boolean; // !Alive
      property Warrior: TWarrior read FWarrior;
      property ThreadCount: integer read FThreadCount;
      property ActiveThread: TThread read GetActiveThread;
    end;
  TThread = class
    protected
      FID: integer;
      FOwner: TProcess;
      FPC: TOperand;
      constructor Create(AOwner: TProcess; APC: TOperand);
    public
      property Process: TProcess read FOwner;
      property PC: TOperand read FPC;
      property ID: integer read FID;
    end;
  TRunState = (
    QUICK_RUN, // graphics should not be drawn
    RUN,  // running, but with some graphics
    RUN_STEP, // execution is pausing between steps for user action
    STOP // game is over
    );
  TMarsStateChangeListener = procedure(const AState: TRunState) of object;
  TMarsThreadListener = procedure(const AThread: TThread) of object;
  TMarsProcessListener = function(const AProcess: TProcess): boolean of object; // a chance to stop the game
  TMarsCoreListener = function(const ACore: TInstruction): boolean of object; // a chance to defer continuation
  TMars = class
    protected
      // main stuff
      FCoreSize: TOperand;
      FCycles, // number of cycles occurred so far
      FMaxCycles,
      FMaxThreads: integer;
      FCore: array of TInstruction;
      FProcesses: array of TProcess;
      FActiveProcess,
      FProcessCount: integer;
      // variables internal state within RunStep used if we have a deferred step due to breakpoint triggering
      FDeferredPC: TOperand;
      FRemove: boolean; // do we die?
      FDefer: boolean; // in a deferred continuation?
      FState: TRunState;
      // listeners
      FStateChangeListeners: array of TMarsStateChangeListener; // listeners for when the state changes
      FWarriorListeners: array of TMarsProcessListener; // listeners for warrior addition and death
      FStepListeners: array of TMarsThreadListener; // listeners for 'before step'
      FCoreListeners: array of TMarsCoreListener; // listeners for all core write events
      procedure NotifyThreadListeners(var AListeners: array of TMarsThreadListener; const AThread: TThread);
      function NotifyWarriorListeners(const AProcess: TProcess): boolean; // return true if any listeners stop the game
      function GetCore(AIndex: TOperand): TInstruction;
      function GetProcess(AIndex: TOperand): TProcess;
      procedure SetState(AState: TRunState);
      procedure Next; // go to next warrior process
      function Checksum: cardinal;
    public
      constructor Create(ACoreSize: TOperand; AMaxCycles, AMaxThreads: integer);
      procedure Clear;
      procedure AddWarrior(AWarrior: TWarrior; AStartPos: TOperand);
      property WarriorCount: integer read FProcessCount;
      property Warrior[AIndex: integer]: TProcess read GetProcess;
      property CORESIZE: TOperand read FCoreSize;
      property MAXTHREADS: integer read FMaxThreads;
      property MAXCYCLES: integer read FMaxCycles;
      property CYCLES: integer read FCycles;
      property Core[AIndex: TOperand]: TInstruction read GetCore;
      property State: TRunState read FState write SetState;
      function Fold(AIndex: TOperand): TOperand;
      function RelativeAddress(AValue: TOperand): integer; overload;
      function RelativeAddress(APC, AValue: TOperand): integer; overload;
      function ActiveProcess: TProcess;
      procedure InitGame;
      procedure Run;
      procedure RunStep;
      // register event listeners
      procedure AddCoreListener(AListener: TMarsCoreListener);
      procedure RemoveCoreListener(AListener: TMarsCoreListener);
      procedure AddStepListener(AListener: TMarsThreadListener);
      procedure RemoveStepListener(AListener: TMarsThreadListener);
      procedure AddWarriorListener(AListener: TMarsProcessListener);
      procedure RemoveWarriorListener(AListener: TMarsProcessListener);
      procedure AddStateChangeListener(AListener: TMarsStateChangeListener);
      procedure RemoveStateChangeListener(AListener: TMarsStateChangeListener);
      // these just register an interest in a particular core location;
      // breakpoints won't be consulted if there is no registered interest in a location;
      // watch counts are culmative; if you add, you should remove when you don't need it
      procedure AddReadWatch(AIndex: TOperand);
      procedure RemoveReadWatch(AIndex: TOperand);
    end;

implementation

{$IFDEF DEBUG_MARS}
uses
  Assembler;

var
  Diag: Text;
{$ENDIF}

const
  EmptyInstruction: TInstruction = (
    Index: -1;
    Owner: nil;
    ContainsCode: False;
    OpCode: oDAT;
    OpModifier: mF;
    A: (AddrMode: aDirect; Value: 0);
    B: (AddrMode: aDirect; Value: 0);
    ReadWatchCount: 0;
    );

(********** TWarrior implementation *******************************)

constructor TWarrior.Create(AID: cardinal; AFilename: string);
begin
  FID := AID;
  FFilename := AFilename;
  FName := '';
  FAuthor := '';
  FHill := '';
  FOrg := 0;
  SetLength(FCore,0);
end;

function TWarrior.GetCore(AIndex: TOperand): TInstruction;
begin
  {$RANGECHECKS ON} // want an exception if it is out of range
  Result := FCore[AIndex];
  {$RANGECHECKS OFF}
end;

function TWarrior.GetName: string;
begin
  Result := FName;
  if (Result = '') then
    Result := FFilename;
  if (Result = '') then
    Result := ConCat('unnamed#',IntToStr(FID));
end;

procedure TWarrior.Append(ACore: TInstruction);
var
  i: integer;
begin
  i := Length;
  SetLength(FCore,Succ(i));
  FCore[i] := ACore;
end;

function TWarrior.Length: TOperand;
begin
  Result := System.Length(FCore);
end;

(********** TProcess implementation *******************************)

constructor TProcess.Create(AWarrior: TWarrior; AMaxThreads: integer);
begin
  FWarrior := AWarrior;
  FAlive := True;
  // assign thread list
  FThreadCount := 0;
  FMaxThreads := AMaxThreads;
  SetLength(FThreads,FMaxThreads);
  FActiveThread := 0;
end;

procedure TProcess.Next; // move the active thread pointer to the next thread
begin
  FActiveThread := (Succ(FActiveThread) mod FThreadCount);
end;

function TProcess.GetActiveThread: TThread; // get the current active thread
begin
  Result := FThreads[FActiveThread];
end;

procedure TProcess.KillThread; // kill the active thread; won't die if it is the last thread!
var i: integer;
  t: TThread;
begin
  if (FThreadCount > 1) then begin
    t := FThreads[FActiveThread];
    for i := FActiveThread to Pred(High(FThreads)) do
      FThreads[i] := FThreads[Succ(i)];
    t.Free;
    Dec(FThreadCount);
    FActiveThread := (FActiveThread mod FThreadCount)
  end else
    FAlive := False;
end;

procedure TProcess.CreateThread(APC: TOperand);
var
  i: integer;
begin
  // too many threads?
  if (FThreadCount >= FMaxThreads) then
    Exit;
  // move all other threads down to make room to insert it
  for i := FThreadCount downto Succ(FActiveThread) do
    FThreads[i] := FThreads[Pred(i)];
  // create the thread
  FThreads[FActiveThread] := TThread.Create(Self,APC);
  // count it
  Inc(FThreadCount);
  // keep active thread active
  FActiveThread := (Succ(FActiveThread) mod FThreadCount);
end;

function TProcess.Dead: boolean; begin Result := not FAlive; end;

(********** TThread implementation *******************************)

constructor TThread.Create(AOwner: TProcess; APC: TOperand);
begin
  FOwner := AOwner;
  FPC := APC;
end;

(********** TMars implementation *******************************)

constructor TMars.Create(ACoreSize: TOperand; AMaxCycles, AMaxThreads: integer);
begin
  FCoreSize := ACoreSize;
  FMaxCycles := AMaxCycles;
  FMaxThreads := AMaxThreads;
  FCycles := 0;
  FDefer := False;
  FState := RUN_STEP;
  // allocate core
  SetLength(FCore,FCoreSize);
  Clear;
  // warrior list
  SetLength(FProcesses,0);
  FActiveProcess := 0;
  FProcessCount := 0;
  // listeners
  SetLength(FCoreListeners,0);
  SetLength(FStepListeners,0);
  SetLength(FWarriorListeners,0);
  SetLength(FStateChangeListeners,0);
end;

function TMars.Checksum: cardinal;
var
  i: integer;
begin
  Result := 0;
  for i := Low(FCore) to High(FCore) do
    Result := (Result + Fold(FCore[i].A.Value) + Fold(FCore[i].B.Value));
end;


procedure TMars.Clear;
var
  i: integer;
begin
  for i := Low(FCore) to High(FCore) do begin
    FCore[i] := EmptyInstruction;
    FCore[i].Index := i;
  end;
end;

function TMars.GetCore(AIndex: TOperand): TInstruction;
begin Result := FCore[Fold(AIndex)]; end;

function TMars.Fold(AIndex: TOperand): TOperand;
begin // probably inefficient at the mo
  Result := AIndex;
  while (Result < 0) do
    Inc(Result,FCoreSize);
  while (Result >= FCoreSize) do
    Dec(Result,FCoreSize);
end;

function TMars.RelativeAddress(AValue: TOperand): integer;
begin
  Result := Fold(AValue);
  if (Result > (CORESIZE div 2)) then
    Result := -(CORESIZE - Result);
end;

function TMars.RelativeAddress(APC, AValue: TOperand): integer; // signed
begin Result := -RelativeAddress(APC - AValue); end;

procedure TMars.AddCoreListener(AListener: TMarsCoreListener);
var
  i: integer;
begin
  i := Length(FCoreListeners);
  SetLength(FCoreListeners,Succ(i));
  FCoreListeners[i] := AListener;
end;

procedure TMars.RemoveCoreListener(AListener: TMarsCoreListener);
var
  i, j: integer;
begin
  // find it
  for i := Low(FCoreListeners) to High(FCoreListeners) do
    if (@FCoreListeners[i] = @AListener) then begin
      j := High(FCoreListeners);
      FCoreListeners[i] := FCoreListeners[j];
      SetLength(FCoreListeners,j); // j is zero-based :-)
      Exit; // done
    end;
end;

procedure TMars.AddWarriorListener(AListener: TMarsProcessListener);
var
  i: integer;
begin
  // add the listener
  i := Length(FWarriorListeners);
  SetLength(FWarriorListeners,Succ(i));
  FWarriorListeners[i] := AListener;
  // inform it of all warriors currently in the MARS
  for i := Low(FProcesses) to High(FProcesses) do
    AListener(FProcesses[i]);
end;

procedure TMars.RemoveWarriorListener(AListener: TMarsProcessListener);
var
  i, j: integer;
begin
  // find it
  for i := Low(FWarriorListeners) to High(FWarriorListeners) do
    if (@FWarriorListeners[i] = @AListener) then begin
      j := High(FWarriorListeners);
      FWarriorListeners[i] := FWarriorListeners[j];
      SetLength(FWarriorListeners,j); // j is zero-based :-)
      Exit; // done
    end;
end;

procedure TMars.AddStepListener(AListener: TMarsThreadListener);
var
  i: integer;
begin
  i := Length(FStepListeners);
  SetLength(FStepListeners,Succ(i));
  FStepListeners[i] := AListener;
end;

procedure TMars.NotifyThreadListeners(var AListeners: array of TMarsThreadListener; const AThread: TThread);
var
  i: integer;
begin
  for i := Low(AListeners) to High(AListeners) do
    AListeners[i](AThread);
end;

function TMars.NotifyWarriorListeners(const AProcess: TProcess): boolean; // return true if any listeners stop the game
var
  i: integer;
begin
  Result := False;
  for i := Low(FWarriorListeners) to High(FWarriorListeners) do
    Result := (Result or FWarriorListeners[i](AProcess));
end;

procedure TMars.RemoveStepListener(AListener: TMarsThreadListener);
var
  i, j: integer;
begin
  // find it
  for i := Low(FStepListeners) to High(FStepListeners) do
    if (@FStepListeners[i] = @FStepListeners) then begin
      j := High(FStepListeners);
      FStepListeners[i] := FStepListeners[j];
      SetLength(FStepListeners,j); // j is zero-based :-)
      Exit; // done
    end;
end;

procedure TMars.AddStateChangeListener(AListener: TMarsStateChangeListener);
var
  i: integer;
begin
  i := Length(FStateChangeListeners);
  SetLength(FStateChangeListeners,Succ(i));
  FStateChangeListeners[i] := AListener;
end;

procedure TMars.RemoveStateChangeListener(AListener: TMarsStateChangeListener);
var
  i, j: integer;
begin
  // find it
  for i := Low(FStateChangeListeners) to High(FStateChangeListeners) do
    if (@FStateChangeListeners[i] = @FStateChangeListeners) then begin
      j := High(FStateChangeListeners);
      FStateChangeListeners[i] := FStateChangeListeners[j];
      SetLength(FStateChangeListeners,j); // j is zero-based :-)
      Exit; // done
    end;
end;

procedure TMars.AddReadWatch(AIndex: TOperand);
begin
  Inc(FCore[Fold(AIndex)].ReadWatchCount);
end;

procedure TMars.RemoveReadWatch(AIndex: TOperand);
begin
  Dec(FCore[Fold(AIndex)].ReadWatchCount);
end;

procedure TMars.SetState(AState: TRunState);
var
  i: integer;
begin
  // set the state
  FState := AState;
  // notify all listeners
  for i := Low(FStateChangeListeners) to High(FStateChangeListeners) do
    FStateChangeListeners[i](FState);
end;

procedure TMars.AddWarrior(AWarrior: TWarrior; AStartPos: TOperand);
var
  i,j: integer;
begin
  FProcessCount := Length(FProcesses);
  SetLength(FProcesses,Succ(FProcessCount));
  FProcesses[FProcessCount] := TProcess.Create(AWarrior,FMaxThreads);
  FProcesses[FProcessCount].CreateThread(Fold(AStartPos + AWarrior.Org));
  for i := 0 to Pred(AWarrior.Length) do begin
    j := Fold(i+AStartPos);
    FCore[j] := AWarrior.Core[i];
    with FCore[j] do begin
      A.Value := Fold(A.Value);
      B.Value := Fold(B.Value);
      Index := j; // restore because it has just been overwritten
    end;
  end;
  // count it
  Inc(FProcessCount);
  // now inform the listeners (after process count is incremented, so that it is accurate if they query the mars for all warriors)
  NotifyWarriorListeners(FProcesses[Pred(FProcessCount)]);
end;

function TMars.GetProcess(AIndex: TOperand): TProcess;
begin
  if (AIndex in [0..Pred(FProcessCount)]) then
    Result := FProcesses[AIndex]
  else
    Result := nil;
end;

function TMars.ActiveProcess: TProcess;
begin Result := GetProcess(FActiveProcess); end;

procedure TMars.Next;
var
  AntiInf: integer;
begin
  AntiInf := FProcessCount;
  repeat
    FActiveProcess := (Succ(FActiveProcess) mod FProcessCount);
    Dec(AntiInf);
  until ActiveProcess.Alive or (AntiInf = 0);
  if AntiInf = 0 then
    FActiveProcess := FProcessCount; // mark it is impossible
end;

procedure TMars.InitGame;
begin
  FCycles := 0;
  NotifyThreadListeners(FStepListeners,ActiveProcess.ActiveThread);
end;

procedure TMars.Run;
var
  i: integer;
begin
  for i := 1 to FMaxCycles do begin
    if (State = STOP) then
      Break;
    RunStep;
  end;
end;

procedure TMars.RunStep;
var
  Temp,
  IR, // current instruction
  IRA, // instruction pointed at by A operand
  IRB: TInstruction; // instruction pointed at by B operand
  TempI,
  PA, // instruction index pointed at by A, relative to PC
  PB, // instruction index pointed at by B, relative to PC
  PC: integer; // program counter for this thread
procedure SetCore(APC: TOperand; var AInstruction: TInstruction); overload;
var
  i,
  ReadWatchCount: integer;
begin
  APC := Fold(APC);
  // remember some 'immutable' fields
  ReadWatchCount := FCore[APC].ReadWatchCount;
  // set it
  FCore[APC] := AInstruction;
  FCore[APC].Owner := ActiveProcess.FWarrior;
  FCore[APC].Index := APC; // restore because we probably overwrote this
  FCore[APC].ReadWatchCount := ReadWatchCount; // restore it
  // inform listeners?
  for i := Low(FCoreListeners) to High(FCoreListeners) do
    FDefer := (FDefer or FCoreListeners[i](FCore[APC])); // remember if anyone defers continuation
end;
type
  TField = (AFIELD,BFIELD);
procedure Add(f: TField; i, APC: TOperand );
var
  j: integer;
begin
  APC := Fold(APC);
  case f of
    AFIELD: FCore[APC].A.Value := Fold(FCore[APC].A.Value + i);
    BFIELD: FCore[APC].B.Value := Fold(FCore[APC].B.Value + i);
  end;
  // set owner
  FCore[APC].Owner := ActiveProcess.FWarrior;
  // inform listeners
  for j := Low(FCoreListeners) to High(FCoreListeners) do
    FDefer := (FDefer or FCoreListeners[j](FCore[APC])); // remember if anyone defers continuation
end;
procedure MoveToNext(APC: TOperand);
var
  InfinityCheck: integer;
begin
  if (ActiveProcess = nil) then
    Exit;
  // thread death?
  if FRemove then begin
    ActiveProcess.KillThread; // auto next
    if ActiveProcess.Dead then // did we kill this warrior?
      if NotifyWarriorListeners(ActiveProcess) then begin // game over???
        FActiveProcess := -1; // flag that it is the end of the game!
        Exit;
      end;
  end else begin // have to increment it to the next instruction
    ActiveProcess.ActiveThread.FPC := Fold(APC);
    ActiveProcess.Next;
  end;
  // swap to the next warrior
  if (FProcessCount = 1) then begin // debugging a single warrior? not end of game until end of cycles then
    if ActiveProcess.Alive then
      InfinityCheck := -1 // an impossible value, so it never triggers a end-of-game
    else begin // hmm, bad news
      FActiveProcess := -1; // flag that it is the end of the debug!
      Exit;
    end;
  end else // we have to check that we aren't the last alive warrior
    InfinityCheck := FActiveProcess; // so we remember where we started
  repeat
    FActiveProcess := (Succ(FActiveProcess) mod FProcessCount);
    if (FActiveProcess = InfinityCheck) then begin // a complete loop of all warriors, and none others are alive?
      FActiveProcess := -1; // flag that it is the end of the game!
      Break;
    end;
  until ActiveProcess.Alive;
  // TODO: check that we haven't only got one warrior left!
  // tell listeners
  if (ActiveProcess <> nil) then
    NotifyThreadListeners(FStepListeners,ActiveProcess.ActiveThread)
  else begin
    SetState(STOP);
    NotifyThreadListeners(FStepListeners,nil);
  end;
end;
begin
  if (State = STOP) then
    Exit;
  if (ActiveProcess = nil) or ActiveProcess.Dead or (FCycles > FMaxCycles) then begin
    SetState(STOP);
    Exit;
  end;
  if FDefer then begin // deferred exec?
    MoveToNext(FDeferredPC);
    FDefer := False;
  end;
  FRemove := False;
  Inc(FCycles);
  PC := ActiveProcess.ActiveThread.FPC;
  IR := GetCore(PC);
  // a field
  PA := IR.A.Value;
  case IR.A.AddrMode of
    aImmediate: begin
      PA := 0;
      IRA := IR;
      end;
    aDirect: begin
      IRA := GetCore(PC + PA);
      end;
    aPreDecA: begin
      Add(AFIELD,-1,PC + PA);
      PA := Fold(PA + GetCore(PC + PA).A.Value); // indirect now?
      IRA := GetCore(PC + PA);
      end;
    aPreDecB: begin
      Add(BFIELD,-1,PC + PA);
      PA := Fold(PA + GetCore(PC + PA).B.Value); // indirect now?
      IRA := GetCore(PC + PA);
      end;
    aPostIncA: begin
      TempI := Fold(PA + GetCore(PC+PA).A.Value);
      IRA := GetCore(PC + TempI);
      Add(AFIELD,1,PC+PA);
      PA := TempI;
      end;
    aPostIncB: begin
      TempI := Fold(PA + GetCore(PC+PA).B.Value);
      IRA := GetCore(PC + TempI);
      Add(BFIELD,1,PC+PA);
      PA := TempI;
      end;
    aIndirectA: begin
      PA := Fold(PA + GetCore(PC + PA).A.Value); // indirect now?
      IRA := GetCore(PC + PA);
      end;
    aIndirectB: begin
      PA := Fold(PA + GetCore(PC + PA).B.Value); // indirect now?
      IRA := GetCore(PC + PA);
      end;
    else
      raise EMars.Create('invalid addressing mode for A');
  end;
  // b field
  PB := IR.B.Value;
  case IR.B.AddrMode of
    aImmediate: begin
      PB := 0;
      IRB := IR;
      end;
    aDirect: begin
      IRB := GetCore(PC + PB);
      end;
    aPreDecA: begin
      Add(AFIELD,-1,PC + PB);
      PB := Fold(PB + GetCore(PC + PB).A.Value); // indirect now?
      IRB := GetCore(PC + PB);
      end;
    aPreDecB: begin
      Add(BFIELD,-1,PC + PB);
      PB := Fold(PB + GetCore(PC + PB).B.Value); // indirect now?
      IRB := GetCore(PC + PB);
      end;
    aPostIncA: begin
      TempI := Fold(PB + GetCore(PC+PB).A.Value);
      IRB := GetCore(PC + TempI);
      Add(AFIELD,1,PC+PB);
      PB := TempI;
      end;
    aPostIncB: begin
      TempI := Fold(PB + GetCore(PC+PB).B.Value);
      IRB := GetCore(PC + TempI);
      Add(BFIELD,1,PC+PB);
      PB := TempI;
      end;
    aIndirectA: begin
      PB := Fold(PB + GetCore(PC + PB).A.Value); // indirect now?
      IRB := GetCore(PC + PB);
      end;
    aIndirectB: begin
      PB := Fold(PB + GetCore(PC + PB).B.Value); // indirect now?
      IRB := GetCore(PC + PB);
      end;
    else
      raise EMars.Create('invalid addressing mode for B');
  end;
{$IFDEF DEBUG_MARS}
  WriteLn(Diag,(FMaxCycles-FCycles+1):6,PC:5,'  ',LowerCase(mnemonic(IR.OpCode)),'.',LowerCase(mnemonic(IR.OpModifier)),
    ' ',mnemonic(IR.A.AddrMode),RelativeAddress(IR.A.Value):5,' , ',mnemonic(IR.B.AddrMode),RelativeAddress(IR.B.Value):5,'  |',
    ActiveProcess.ThreadCount:4,', d ',IRA.Index:4,',',IRB.Index:4,' a ',Fold(IRA.A.Value):4,',',Fold(IRA.B.Value):4,
    ' b ',Fold(IRB.A.Value):4,',',Fold(IRB.B.Value):4,' | ',Checksum:6);
{$ENDIF}
  // command itself
  case IR.OpCode of
    oDAT: FRemove := True; // kill!
    oMOV: begin
      Temp := GetCore(PC + PB);
      case IR.OpModifier of
        mA:  Temp.A.Value := IRA.A.Value;
        mB:  Temp.B.Value := IRA.B.Value;
        mAB: Temp.B.Value := IRA.A.Value;
        mBA: Temp.A.Value := IRA.B.Value;
        mF: begin
             Temp.A.Value := IRA.A.Value;
             Temp.B.Value := IRA.B.Value;
          end;
        mX: begin
             Temp.A.Value := IRA.B.Value;
             Temp.B.Value := IRA.A.Value;
          end;
        mI:  Temp := IRA;
        else
          raise EMars.Create('Bad MOV OpModifier');
      end;
      SetCore(PC + PB,Temp);
      end;
    oADD: begin
      Temp := GetCore(PC + PB);
      case IR.OpModifier of
        mA:  Temp.A.Value := Fold(IRB.A.Value + IRA.A.Value);
        mB:  Temp.B.Value := Fold(IRB.B.Value + IRA.B.Value);
        mBA: Temp.A.Value := Fold(IRB.A.Value + IRA.B.Value);
        mAB: Temp.B.Value := Fold(IRB.B.Value + IRA.A.Value);
        mI,
        mF: begin
             Temp.A.Value := Fold(IRB.A.Value + IRA.A.Value);
             Temp.B.Value := Fold(IRB.B.Value + IRA.B.Value);
          end;
        mX: begin
             Temp.A.Value := Fold(IRB.A.Value + IRA.B.Value);
             Temp.B.Value := Fold(IRB.B.Value + IRA.A.Value);
          end;
        else
          raise EMars.Create('Bad ADD OpModifier');
      end;
      SetCore(PC + PB,Temp);
      end;
    oSUB: begin
      Temp := GetCore(PC + PB);
      case IR.OpModifier of
        mA:  Temp.A.Value := Fold(IRB.A.Value - IRA.A.Value);
        mB:  Temp.B.Value := Fold(IRB.B.Value - IRA.B.Value);
        mBA: Temp.A.Value := Fold(IRB.A.Value - IRA.B.Value);
        mAB: Temp.B.Value := Fold(IRB.B.Value - IRA.A.Value);
        mI,
        mF: begin
             Temp.A.Value := Fold(IRB.A.Value - IRA.A.Value);
             Temp.B.Value := Fold(IRB.B.Value - IRA.B.Value);
          end;
        mX: begin
             Temp.A.Value := Fold(IRB.A.Value - IRA.B.Value);
             Temp.B.Value := Fold(IRB.B.Value - IRA.A.Value);
          end;
        else
          raise EMars.Create('Bad SUB OpModifier');
      end;
      SetCore(PC + PB,Temp);
      end;
    oMUL: begin
      Temp := GetCore(PC + PB);
      case IR.OpModifier of
        mA:  Temp.A.Value := Fold(IRB.A.Value * IRA.A.Value);
        mB:  Temp.B.Value := Fold(IRB.B.Value * IRA.B.Value);
        mBA: Temp.A.Value := Fold(IRB.A.Value * IRA.B.Value);
        mAB: Temp.B.Value := Fold(IRB.B.Value * IRA.A.Value);
        mI,
        mF: begin
             Temp.A.Value := Fold(IRB.A.Value * IRA.A.Value);
             Temp.B.Value := Fold(IRB.B.Value * IRA.B.Value);
          end;
        mX: begin
             Temp.A.Value := Fold(IRB.A.Value * IRA.B.Value);
             Temp.B.Value := Fold(IRB.B.Value * IRA.A.Value);
          end;
        else
          raise EMars.Create('Bad MUL OpModifier');
      end;
      SetCore(PC + PB,Temp);
      end;
    oDIV: begin
      Temp := GetCore(PC + PB);
      case IR.OpModifier of
        mA: if (IRA.A.Value <> 0) then // div0 check
              Temp.A.Value := Fold(IRB.A.Value div IRA.A.Value)
            else
              FRemove := True;
        mB: if (IRA.B.Value <> 0) then // div0 check
              Temp.B.Value := Fold(IRB.B.Value div IRA.B.Value)
            else
              FRemove := True;
        mBA:if (IRA.B.Value <> 0) then // div0 check
              Temp.A.Value := Fold(IRB.A.Value div IRA.B.Value)
            else
              FRemove := True;
        mAB:if (IRA.A.Value <> 0) then // div0 check
              Temp.B.Value := Fold(IRB.B.Value div IRA.A.Value)
            else
              FRemove := True;
        mI,
        mF: begin
              // note that even if one fails, the other has a chance to happen
              if (IRA.A.Value <> 0) then // div0 check
                Temp.A.Value := Fold(IRB.A.Value div IRA.A.Value)
              else
                FRemove := True;
              if (IRA.B.Value <> 0) then // div0 check
                Temp.B.Value := Fold(IRB.B.Value div IRA.B.Value)
              else
                FRemove := True;
          end;
        mX: begin
              // note that even if one fails, the other has a chance to happen
              if (IRA.B.Value <> 0) then // div0 check
                Temp.A.Value := Fold(IRB.A.Value div IRA.B.Value)
              else
                FRemove := True;
              if (IRA.A.Value <> 0) then // div0 check
                Temp.B.Value := Fold(IRB.B.Value div IRA.A.Value)
              else
                FRemove := True;
          end;
        else
          raise EMars.Create('Bad MUL OpModifier');
      end;
      SetCore(PC + PB,Temp);
      end;
    oMOD: begin
      Temp := GetCore(PC + PB);
      case IR.OpModifier of
        mA: if (IRA.A.Value <> 0) then // div0 check
              Temp.A.Value := Fold(IRB.A.Value mod IRA.A.Value)
            else
              FRemove := True;
        mB: if (IRA.B.Value <> 0) then // div0 check
              Temp.B.Value := Fold(IRB.B.Value mod IRA.B.Value)
            else
              FRemove := True;
        mBA:if (IRA.B.Value <> 0) then // div0 check
              Temp.A.Value := Fold(IRB.A.Value mod IRA.B.Value)
            else
              FRemove := True;
        mAB:if (IRA.A.Value <> 0) then // div0 check
              Temp.B.Value := Fold(IRB.B.Value mod IRA.A.Value)
            else
              FRemove := True;
        mI,
        mF: begin
              // note that even if one fails, the other has a chance to happen
              if (IRA.A.Value <> 0) then // div0 check
                Temp.A.Value := Fold(IRB.A.Value mod IRA.A.Value)
              else
                FRemove := True;
              if (IRA.B.Value <> 0) then // div0 check
                Temp.B.Value := Fold(IRB.B.Value mod IRA.B.Value)
              else
                FRemove := True;
          end;
        mX: begin
              // note that even if one fails, the other has a chance to happen
              if (IRA.B.Value <> 0) then // div0 check
                Temp.A.Value := Fold(IRB.A.Value mod IRA.B.Value)
              else
                FRemove := True;
              if (IRA.A.Value <> 0) then // div0 check
                Temp.B.Value := Fold(IRB.B.Value mod IRA.A.Value)
              else
                FRemove := True;
          end;
        else
          raise EMars.Create('Bad MUL OpModifier');
      end;
      SetCore(PC + PB,Temp);
      end;
    oJMP:
      PC := PC + PA - 1; // will fold later
    oJMZ: begin
      // will fold PC later
      case IR.OpModifier of
        mA, mBA: if (IRB.A.Value = 0) then PC := PC + PA - 1;
        mB, mAB: if (IRB.B.Value = 0) then PC := PC + PA - 1;
        mF, mX, mI: if (IRB.A.Value = 0) and (IRB.B.Value = 0) then PC := PC + PA - 1;
      end;
      end;
    oJMN: begin
      // will fold PC later
      case IR.OpModifier of
        mA, mBA: if (IRB.A.Value <> 0) then PC := PC + PA - 1;
        mB, mAB: if (IRB.B.Value <> 0) then PC := PC + PA - 1;
        mF, mX, mI: if (IRB.A.Value <> 0) or (IRB.B.Value <> 0) then PC := PC + PA - 1;
      end;
      end;
    oDJN: begin
      TempI := Fold(PC + PB); // remember this just in case PC changes
      Temp := GetCore(PC + PB);
      // will fold PC later
      case IR.OpModifier of
        mA, mBA: begin
          Temp.A.Value := Fold(Temp.A.Value - 1);
          if (IRB.A.Value <> 1) then
            PC := PC + PA - 1;
          end;
        mB, mAB: begin
          Temp.B.Value := Fold(Temp.B.Value - 1);
          if (IRB.B.Value <> 1) then
            PC := PC + PA - 1;
          end;
        mF, mX, mI: begin
          Temp.A.Value := Fold(Temp.A.Value - 1);
          Temp.B.Value := Fold(Temp.B.Value - 1);
          if (IRB.A.Value <> 1) or (IRB.B.Value <> 1) then
            PC := PC + PA - 1;
          end;
      end;
      SetCore(TempI,Temp);
      end;
    oSEQ: begin
      case IR.OpModifier of
        mA:  if (IRA.A.Value = IRB.A.Value) then Inc(PC);
        mB:  if (IRA.B.Value = IRB.B.Value) then Inc(PC);
        mAB: if (IRA.A.Value = IRB.B.Value) then Inc(PC);
        mBA: if (IRA.B.Value = IRB.A.Value) then Inc(PC);
        mF:  if (IRA.A.Value = IRB.A.Value) and (IRA.B.Value = IRB.B.Value) then Inc(PC);
        mX:  if (IRA.A.Value = IRB.B.Value) and (IRA.B.Value = IRB.A.Value) then Inc(PC);
        mI:  if (IRA.OpCode = IRB.OpCode) and (IRA.OpModifier = IRB.OpModifier) and
                (IRA.A.AddrMode = IRB.A.AddrMode) and (IRA.A.Value = IRB.A.Value) and
                (IRA.B.AddrMode = IRB.B.AddrMode) and (IRA.B.Value = IRB.B.Value) then Inc(PC);
      end;
      end;
    oSNE: begin
      case IR.OpModifier of
        mA:  if (IRA.A.Value <> IRB.A.Value) then Inc(PC);
        mB:  if (IRA.B.Value <> IRB.B.Value) then Inc(PC);
        mAB: if (IRA.A.Value <> IRB.B.Value) then Inc(PC);
        mBA: if (IRA.B.Value <> IRB.A.Value) then Inc(PC);
        mF:  if (IRA.A.Value <> IRB.A.Value) or (IRA.B.Value <> IRB.B.Value) then Inc(PC);
        mX:  if (IRA.A.Value <> IRB.B.Value) or (IRA.B.Value <> IRB.A.Value) then Inc(PC);
        mI:  if (IRA.OpCode <> IRB.OpCode) or (IRA.OpModifier <> IRB.OpModifier) or
                (IRA.A.AddrMode <> IRB.A.AddrMode) or (IRA.A.Value <> IRB.A.Value) or
                (IRA.B.AddrMode <> IRB.B.AddrMode) or (IRA.B.Value <> IRB.B.Value) then Inc(PC);
      end;
      end;
    oSLT: begin
      case IR.OpModifier of
        mA:  if (IRA.A.Value < IRB.A.Value) then Inc(PC);
        mB:  if (IRA.B.Value < IRB.B.Value) then Inc(PC);
        mAB: if (IRA.A.Value < IRB.B.Value) then Inc(PC);
        mBA: if (IRA.B.Value < IRB.A.Value) then Inc(PC);
        mF, mI: if (IRA.A.Value < IRB.A.Value) and (IRA.B.Value < IRB.B.Value) then Inc(PC);
        mX:  if (IRA.A.Value < IRB.B.Value) and (IRA.B.Value < IRB.A.Value) then Inc(PC);
      end;
      end;
    oSPL:
      with ActiveProcess do begin
        CreateThread(Succ(PC));
        Inc(PC,PA-1);
      end;
    oNOP: begin
      end;
    else
      raise EMars.CreateFmt('unsupported opcode %s',[mnemonic(IR.OpCode)]);
  end;
  if FDefer then // remember new pointer?
    FDeferredPC := Succ(PC)
  else // no deferred continuation?
    MoveToNext(Succ(PC));
end;

{$IFDEF DEBUG_MARS}
initialization
  AssignFile(Diag,'mymars.dbg');
  Rewrite(Diag);
finalization
  WriteLn(Diag,'1 0'); // score to match the exhaust printout
  CloseFile(Diag);
{$ENDIF}
end.
