unit Parser;

(***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************)

interface

uses
  Mars,
  Assembler,
  SysUtils;

type
  EParse = class(Exception);

{ ParseRed: read one or more warriors from a .red assembly file.  This is explained in
  the documentation that accompanies pmars
  Cravats:
  * Whitespace is any character with an ASCII code of less than or equal to space (32), e.g. " " and tabs
  * FULL-LINE-COMMENTS can only be preceded on the line by whitespace
  * only FULL-LINE-COMMENTS are scanned for special directives;
    reasoning: otherwise quoting a ;redcode preprocessor directive
    in the ignoreable description would trigger parsing to start
  * A line that ends with a '\' character continues to the next line.  The \ character is remoevd
    before parsing the meta-line
  * Directives and labels are case-insenstive
}
function ParseRed(Mars: TMars; AFilename: string; ADiagnosticsFilename: string = ''): TWarrior;

{ a basic reader of preprocessed redcode files, which expect to have the extension .rc
  format is fixed }
function ParseRC(AID: cardinal; AFilename: string): TWarrior;

implementation

uses
  StrUtils;

(********* ParseRed implementation *******************************************)

function ParseRed(Mars: TMars; AFilename: string; ADiagnosticsFilename: string = ''): TWarrior;
type
  TLabel = record
    Name: string;
    Value: TOperand;
    Resolved,
    Standard: boolean;
    end;
  TToken = record
    end;
var
  Labels: array of TLabel;
function GetLabel(AName: string): integer; // returns index; -1 if not found
begin
  for Result := Low(Labels) to High(Labels) do
    if SameText(Labels[Result].Name,AName) then
      Exit;
  Result := -1; // not found
end;
function AddLabel(AName: string; AValue: TOperand = 0; AResolved: boolean = false; AStandard: boolean = false): integer; // returns index
begin
  Result := GetLabel(AName);
  if AResolved then
    if (Result <> -1) then // dup?
      // check it isn't already resolved
      if Labels[Result].Resolved then
        raise EParse.CreateFmt('Label "%s" is already resolved',[AName]);
  if (Result <> -1) then begin // need to add it
    Result := Length(Labels);
    SetLength(Labels,Succ(Result));
  end;
  // need to set it
  with Labels[Result] do begin
    Name := AName;
    Value := AValue;
    Resolved := AResolved;
    Standard := AStandard;
  end;
end;
procedure InitStandardLabels;
begin
  AddLabel('CORESIZE',Mars.CORESIZE,true,true);
  AddLabel('MAXPROCESSES',Mars.MAXTHREADS,true,true);
  AddLabel('MAXCYCLES',Mars.MAXCYCLES,true,true);
//AddLabel('MAXLENGTH         value of -l parameter (default: 100)
//AddLabel('MINDISTANCE       value of -d parameter (default: 100)
//AddLabel('ROUNDS            (+) value of -r parameter (default: 1)
//AddLabel('PSPACESIZE        (+) value of -S parameter (default: 1/16th CORESIZE)
//AddLabel('CURLINE           current line in generated assembly (starts with 0)
//AddLabel('VERSION           pMARS version ("60" is v0.6.0)
//AddLabel('WARRIORS          number of warriors specified on command lineend;
end;
begin
  InitStandardLabels;
end;


(********* ParseRC implementation *******************************************)

function ParseRC(AID: cardinal; AFilename: string): TWarrior;
var
  sLine, // full line
  sLbl, // label
  sOpCode,
  sOpModifier,
  sValueA,
  sValueB: string;
  cAddrModeA,
  cAddrModeB: char;
  sOrg: string;
  CurLine: integer;
  Core: TInstruction;
  F: Text;
procedure ParseLine;
begin
  sLbl := Trim(Copy(sLine,1,7));
  sOpCode := Copy(sLine,8,3);
  sOpModifier := Copy(sLine,12,2);
  if Length(sLine) >= 15 then
    cAddrModeA := sLine[15]
  else
    Exit;
  sValueA := Trim(Copy(sLine,16,6));
  if Length(sLine) >= 24 then
    cAddrModeB := sLine[24]
  else
    Exit;
  sValueB := Trim(Copy(sLine,25,6));
end;
function ReadLine: boolean;
begin
  repeat
    Result := not EOF(F);
    if not Result then
      Exit;
    ReadLn(F,sLine);
    ParseLine;
  until (Trim(sOpCode) <> '');
end;
begin
  Result := TWarrior.Create(AID,AFilename);
  AssignFile(F,AFilename);
  Reset(F);
  sOrg := 'impossble value so that default is used';
  CurLine := 0;
  while ReadLine do begin
    if (sOpCode = 'ORG') then begin
      sOrg := sValueA;
      Continue;
    end;
    if (sLbl = sOrg) then
      Result.Org := CurLine;
    with Core do begin
      Owner := Result;
      OpCode := Assembler.OpCode(sOpCode);
      ContainsCode := (OpCode <> oDAT);
      OpModifier := Assembler.OpModifier(sOpModifier);
      A.AddrMode := Assembler.AddrMode(cAddrModeA);
      A.Value := Assembler.Value(sValueA);
      B.AddrMode := Assembler.AddrMode(cAddrModeB);
      B.Value := Assembler.Value(sVAlueB);
    end;
    Result.Append(Core);
    Inc(CurLine);
  end;
  CloseFile(F);
end;

end.
 
