unit BreakpointList;

(***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************)

interface

uses
  Mars,
  Breakpoint,
  CoreDetails,
  BreakpointEditor,
  Classes,
  Types,
  StdCtrls,
  Controls;

type
  TBreakpointList = class(TCustomControl)
    protected
      FBreakpoints: TBreakpointManager;
      FCoreDetails: TCoreDetails;
      FEditor: TFrmEditBreakpoint;
      FList: TListbox;
      procedure OnListClick(Sender: TObject); // when someone clicks on a breakpoint in the list
      procedure OnListDblClick(Sender: TObject); // when someone double-clicks on a breakpoint in the list
      procedure OnChange(Sender: TBreakpointManager); // my listener
      procedure DrawBreakpoint(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState);
      procedure Sync;
    public
      constructor Create(ABreakpoints: TBreakpointManager; AParent: TWinControl; AOwner: TComponent); reintroduce;
      property CoreDetails: TCoreDetails read FCoreDetails write FCoreDetails;
      destructor Destroy; override;
    end;

implementation

uses
  Graphics,
  Windows,
  SysUtils, Forms;

(***** TBreakpointList implementation *******************************)

constructor TBreakpointList.Create(ABreakpoints: TBreakpointManager; AParent: TWinControl; AOwner: TComponent);
begin
  inherited Create(AOwner);
  Parent := AParent;
  FCoreDetails := nil;
  // make breakpoint change listener
  FBreakpoints := ABreakpoints;
  FBreakpoints.AddBreakpointListener(OnChange);
  // graphical bits
  FList := TListBox.Create(Self);
  with FList do begin
    Parent := Self;
    Align := alClient;
    Style := lbOwnerDrawFixed;
    OnDrawItem := DrawBreakpoint;
    OnClick := OnListClick;
    OnDblClick := OnListDblClick;
  end;
  // editor
  FEditor := TFrmEditBreakpoint.Create(Owner);
  with FEditor do begin
    Parent := Self;
  end;
  // show us
  Sync;
end;

destructor TBreakpointList.Destroy;
begin
  FBreakpoints.RemoveBreakpointListener(OnChange);
  inherited;
end;

procedure TBreakpointList.OnListClick(Sender: TObject); // when someone clicks on a breakpoint in the list
var
  Bp: TBreakpoint;
  Index: TOperand;
begin
  if Assigned(FCoreDetails) and (FList.ItemIndex > -1) then begin
    Bp := TBreakpoint(FList.Items.Objects[FList.ItemIndex]);
    if (Bp.Trigger = BRK_PC) then
      Index := Bp.PC
    else if (Bp.Trigger = BRK_WRITE) then
      Index := Bp.Addr
    else // await the day when we have more trigger types
      Exit;
    FCoreDetails.Focus(Index);
  end;
end;

procedure TBreakpointList.OnListDblClick(Sender: TObject); // when someone double-clicks on a breakpoint in the list
begin
  with FEditor do begin
    Caption := 'Edit Breakpoint';
    if ShowModal = mrOK then
      // set it
      ;
  end;
end;

procedure TBreakpointList.DrawBreakpoint(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState);
const
  GUTTER = 15;
var
  Bp: TBreakpoint;
  s: string;
begin
  s := FList.Items[Index];
  Bp := TBreakpoint(FList.Items.Objects[Index]);
  with (Control as TListbox).Canvas do begin
    // triggered?
    if Bp.Triggered then begin
      Brush.Color := clRed;
      Font.Color := clYellow;
    end else begin
      Brush.Color := clWindow;
      Font.Color := clWindowText;
    end;
    // fill background
    FillRect(Rect);
    // text
    TextRect(Rect,Rect.Left+GUTTER,Rect.Top,s);
    // selected?
    if (odSelected in State) then begin
      Brush.Color := clActiveCaption;
      Inc(Rect.Left);
      Inc(Rect.Top);
      Dec(Rect.Right);
      Dec(Rect.Bottom);
      FrameRect(Rect);
    end;
  end;
  // draw background
end;

procedure TBreakpointList.OnChange(Sender: TBreakpointManager);
begin
  Sync;
end;

procedure TBreakpointList.Sync;
var
  i, first: integer;
  Bp: TBreakpoint;
  s: string;
begin
  // sync it all
  FList.Clear;
  First := 0;
  for i := 0 to Pred(FBreakpoints.Count) do begin
    Bp := FBreakpoints.Breakpoint[i];
    // work out label
    if Assigned(Bp.Warrior) then begin
      s := bp.Warrior.Name;
      if (s = '') then
        s := Bp.Warrior.Filename;
    end else
      s := 'any warrior';
    s := ConCat('When ',s,' ');
    case Bp.Trigger of
      BRK_PC: begin
        s := ConCat(s,'executes ',IntToStr(Bp.PC));
        end;
      BRK_WRITE: begin
        s := ConCat(s,'sets ');
        if (Bp.Field in [BRK_AFIELD,BRK_BOTHFIELDS]) then
          s := ConCat(s,'the A Operand to ',IntToStr(Bp.ValueA));
        if (Bp.Field = BRK_BOTHFIELDS) then
          s := ConCat(s,' and sets ');
        if (Bp.Field in [BRK_BFIELD,BRK_BOTHFIELDS]) then
          s := ConCat(s,'the B Operand to ',IntToStr(Bp.ValueB));
        if (Bp.Field = BRK_NA) then
          s := ConCat(s,'any field');
        s := ConCat(s,' at ',IntToStr(Bp.Addr));
        end;
    end;
    if Bp.Triggered then
      if (First = 0) then // not set?
        First := i; // remember it
    // add it
    FList.AddItem(s,Bp);
  end;
  if (FList.Count > 0) then begin // not an empty list?
    FList.TopIndex := First; // ensure it is visible
    FList.ItemIndex := First;
  end;
  FList.Repaint;
end;

end.
