unit CoreDetails;

(***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************)

interface

uses
  Mars,
  ExplainAddressing,
  Classes,
  Controls;

type
  TCoreDetails = class(TCustomControl)
    protected
      FPC, // the currently thought PC (as drawn, not necessarily synced with the mars until after an event
      FTop: TOperand; // the index of the instruction at the top of the window
      FSeperateFocus: boolean; // a non-PC focus?
      FEcho, // is it showing a debug at the moment?
      FExplained: boolean; // is the FExplain relevent?
      FExplain: TExplaination;
      FFocus, // the currently middle instruction (independant of FPC)
      FLines, // number of visible lines
      FAddrChars, // number of characters in the gutter
      FGutterWidth, // width of the gutter which displays the line numbers etc
      FAField, FBField, // offset of the start of these fields
      FLineHeight: integer; // the height of a line at the moment
      FMars: TMars;
      FThread: Mars.TThread;
      // drawing suboperations
      procedure PaintScroll;
      procedure DrawCursor;
      procedure DrawHidden;
      procedure DrawLineNum(Y: Integer; i: TOperand);
      procedure DrawLine(Y: Integer; i: TOperand);
      procedure DrawExplaination;
      // events
      function OnCoreChange(const ACore: TInstruction): boolean;
      procedure OnStep(const AThread: Mars.TThread);
      procedure HandleResize(Sender: TObject);
      procedure OnStateChange(const AState: TRunState);
      // involved getters
      function GetPC: TOperand;
      // setters
      procedure SetThread(AThread: Mars.TThread);
    public
      constructor Create(AMars: TMars; AOwner: TComponent); reintroduce;
      property ActiveThread: Mars.TThread read FThread write SetThread;
      property Rows: integer read FLines;
      property PC: TOperand read GetPC;
      procedure Paint; override;
      procedure Focus(AIndex: TOperand); // makes this index the centre of the view; does not affect PC etc
    end;

implementation

uses
  Assembler,
  GlobalConfig,
  Graphics,
  Windows,
  Math,
  SysUtils;

(******* TCoreDetails implementation ************************************)

constructor TCoreDetails.Create(AMars: TMars; AOwner: TComponent);
const
  EM = 12;
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque];
  DoubleBuffered := GlobalConfig.DoubleBuffered;
  FMars := AMars;
  FPC := -1000; // impossible value to trigger a full repaint the first time
  FEcho := True;
  FExplained := False;
  FSeperateFocus := True;
  FFocus := 0;
  // dimensions
  with Canvas.Font do begin
    Name := 'Courier New';
    Size := 11;
    FLineHeight := Abs(Height)+2;
    FAddrChars := Length(IntToStr(FMars.CORESIZE))+1;
    FGutterWidth := EM*FAddrChars+1;
    FAField := FGutterWidth + (FAddrChars*EM);
    FBField := FAField + FAddrChars*EM+2;
  end;
  // listeners
  OnResize := HandleResize;
  FMars.AddCoreListener(OnCoreChange);
  FMars.AddStepListener(OnStep);
  FMars.AddStateChangeListener(OnStateChange);
end;

function TCoreDetails.OnCoreChange(const ACore: TInstruction): boolean;
var
  Y: integer;
  r: TRect;
begin
  Result := False;
  // disable echo?
  if not FEcho then
    Exit;
  // where are we talking about?
  Y := FMars.Fold(ACore.Index + FTop) * FLineHeight;
  // is it visible?
  if (Y >= 0) and ((Y - FLineHeight) <= Height) then begin
    r := Rect(0,Y,Width,Y+FLineHeight);
    InvalidateRect(Handle,@r,False);
    Update;
  end;
end;

procedure TCoreDetails.OnStep(const AThread: Mars.TThread);
begin
  // disabled echo?
  if not FEcho then
    Exit;
  SetThread(AThread);
  FExplained := False; //(State.RunState = RUN_STEP);
  if FExplained then
    ExplainAddressing.Explain(FMars,FPC,FExplain);
{ FThread := AThread;
  HandleResize(Self); // calculate new top
  Repaint; // update view }
end;

procedure TCoreDetails.DrawHidden;
begin
  with Canvas do begin
    Brush.Color := clDkGray;
    FillRect(ClientRect);
  end;
end;

procedure TCoreDetails.OnStateChange(const AState: TRunState);
begin
  FEcho := (AState in [RUN_STEP,STOP]);
  if not FEcho then
    DrawHidden
  else begin
    // ensure we are showing where we want to be
    Invalidate; // ensure we don't have any partial messing around
    FPC := -1;
    if (FMars.ActiveProcess <> nil) then
      SetThread(FMars.ActiveProcess.ActiveThread)
    else
      SetThread(nil);
  end;
end;

procedure TCoreDetails.SetThread(AThread: Mars.TThread);
begin
  FThread := AThread;
  if Assigned(FThread) then begin // check for common condition
    if (FThread.PC = Succ(FPC)) then begin // just moved the BP down one line?
      PaintScroll; // fudge it
    end else
      Invalidate; // full repaint needed
    FPC := FThread.PC;
    FFocus := FPC;
    FSeperateFocus := False;
    HandleResize(Self); // recalculate the FTop
  end else begin
    FEcho := False;
    Invalidate; // full repaint needed
  end;
  Update; // update view
end;

function TCoreDetails.GetPC: TOperand;
begin
  if Assigned(FThread) then
    Result := FThread.PC
  else // fudge it
    Result := 0;
end;

procedure TCoreDetails.Focus(AIndex: TOperand); // makes this index the centre of the view; does not affect PC etc
var
  Half: integer;
begin
  FFocus := AIndex;
  Half := FLines div 2;
  FTop := FMars.Fold(FFocus - Half);
  FSeperateFocus := True;
  Repaint;
end;

procedure TCoreDetails.HandleResize(Sender: TObject);
begin
  FLines := Ceil(Height / FLineHeight);
  FTop := FFocus - (FLines div 2);
end;

procedure TCoreDetails.PaintScroll;
var
  i: integer;
  Src, Dest: TRect;
begin
  if not Visible then
    Exit;
  Inc(FTop); // scroll over top line
  Dest := Rect(FGutterWidth,1,Width,Pred(FLineHeight));
  Src := Dest;
  for i := 2 to (FLines-2) do begin
    Inc(Src.Top,FLineHeight);
    Inc(Src.Bottom,FLineHeight);
    Canvas.CopyRect(Dest,Canvas,Src);
    Dest := Src;
  end;
  // block out and repaint the cursor
  Canvas.FillRect(Rect(FGutterWidth-FLineHeight-2,1,FGutterWidth-1,Height-1));
  DrawCursor;
  // the tricky bit at the bottom, the next new line
  for i := (FLines - 2) to Pred(FLines) do
    DrawLine(i*FLineHeight,i);
  // the border we might have overwritten
  with Canvas do begin
    Brush.Color := clBtnShadow;
    MoveTo(Pred(FGutterWidth),Pred(Height));
    LineTo(Pred(Width),Pred(Height));
    LineTo(Pred(Width),Height-(FLineHeight*3));
    Brush.Color := clWindow;
  end;
end;

procedure TCoreDetails.DrawLineNum(Y: Integer; i: TOperand);
var
  s: string;
begin
  with Canvas do begin
    // draw linenumber
    Font.Color := clLtGray;
    Str(FMars.RelativeAddress(PC,FTop + i):FAddrChars,s);
    TextRect(Rect(1,Y,FGutterWidth-1,Y+Pred(FLineHeight)),2,Y,s);
  end;
end;

procedure TCoreDetails.DrawExplaination;
var
  X: array[ExplainAddressing.TExplainField] of integer;
  i: integer;
  OldPenWidth: integer;
  OldPenColor, OldBrushColor: TColor;
  OldPenStyle: TPenStyle;
begin
  if not FExplained then
    Exit;
  X[efINST] := FGutterWidth;
  X[efAFIELD] := FAField;
  X[efBFIELD] := FBField;
  with FExplain, Canvas do begin
    OldPenWidth := Pen.Width;
    OldPenColor := Pen.Color;
    OldPenStyle := Pen.Style;
    OldBrushColor := Brush.Color;
    Pen.Width := 3;
    Pen.Color := clMaroon;
    Brush.Color := clWindow;
    for i := Low(Arrows) to High(Arrows) do
      with Arrows[i] do begin
        if (Line = elDIRECT) then
          Pen.Style := psDash
        else
          Pen.Style := psDot;
        MoveTo(X[Src.Field],(Src.Idx-FTop)*FLineHeight);
        LineTo(X[Dest.Field],(Dest.Idx-FTop)*FLineHeight);
      end;
    Pen.Width := OldPenWidth;
    Pen.Color := OldPenColor;
    Pen.Style := OldPenStyle;
    Brush.Color := OldBrushColor;
  end;
end;

procedure TCoreDetails.DrawCursor;
var
  Y,
  i: integer;
  OldColour: TColor;
begin
  // current process counter
  i := FMars.Fold(FPC - FTop);
  if (i >= 0) and (i <= FLines) then begin
    Y := i * FLineHeight;
    with Canvas do begin
      // set colour
      if Assigned(FThread) then
        Brush.Color := FThread.Process.Warrior.ID // not necessarily the owner of this cell..
      else
        Brush.Color := clLtGray;
      // draw cursor
      Polygon([Point(FGutterWidth-FLineHeight,Y+1),Point(FGutterWidth-2,Y+(FLineHeight div 2)),
        Point(FGutterWidth-FLineHeight,Y+FLineHeight-2)]);
      // restore colour for normal backgrounds
      Brush.Color := clWindow;
    end;
  end;
  // focus?
  if FSeperateFocus then begin
    i := FMars.Fold(FFocus - FTop);
    if (i >= 0) and (i <= FLines) then begin
      Y := i * FLineHeight;
      with Canvas do begin
        // set colour
        OldColour := Pen.Color;
        Brush.Color := clLime;
        Pen.Color := clGreen;
        // draw cursor
        Polygon([Point(FGutterWidth-FLineHeight+3,Y+(FLineHeight div 5)),Point(FGutterWidth-3,Y+(FLineHeight div 2)),
          Point(FGutterWidth-FLineHeight+3,Y+FLineHeight-(FLineHeight div 5))]);
        // restore colour for normal backgrounds
        Brush.Color := clWindow;
        Pen.Color := OldColour;
      end;
    end;
  end;
end;

procedure TCoreDetails.DrawLine(Y: Integer; i: TOperand);
var
  s,a,b: string;
  Core: TInstruction;
begin
 with Canvas do begin
  // draw instruction
  // what colour and style?
  Core := FMars.Core[FTop + i];
  if Assigned(Core.Owner) then
    Font.Color := Core.Owner.ID
  else
    Font.Color := clGray;
  if Core.ContainsCode then
    Font.Style := [fsBold]
  else
    Font.Style := [];
  // what to write?
  Str(FMars.RelativeAddress(Core.A.Value):FAddrChars,a);
  Str(FMars.RelativeAddress(Core.B.Value):FAddrChars,b);
  s := ConCat(Assembler.mnemonic(Core.OpCode),'.',Assembler.mnemonic(Core.OpModifier),' ',
    Assembler.mnemonic(Core.A.AddrMode),a,' ',Assembler.mnemonic(Core.B.AddrMode),b);
  // draw it
  TextRect(Rect(Succ(FGutterWidth),Y,Left+Width,Y+Pred(FLineHeight)),FGutterWidth+2,Y,s);
  // restore font if it changed
  if Core.ContainsCode then
    Font.Style := [];
 end;
end;

procedure TCoreDetails.Paint;
var
  i: TOperand;
  Y: integer; // pixel offset of top of current line
begin
  Y := 1;
  if not FEcho then
    Exit;
  with Canvas do begin
    // clear background
    Brush.Color := clWindow;
    FillRect(ClipRect);
    // draw the cells
    for i := 0 to FLines do begin
      // visible?
      if ((Y >= ClipRect.Top) and (Y <= ClipRect.Bottom)) then begin
        // linenumber visible?
       if (ClipRect.Left < FGutterWidth) then
        DrawLineNum(Y,i);
       // draw the instruction
       DrawLine(Y,I);
      end;
      // next line
      Inc(Y,FLineHeight);
    end;
    DrawCursor;
    DrawExplaination;
    // draw border
    Brush.Color := clBtnShadow;
    MoveTo(0,ClientHeight-1);
    LineTo(0,0);
    LineTo(ClientWidth-1,0);
    Brush.Color := clBtnHighlight;
    LineTo(ClientWidth-1,ClientHeight-1);
    LineTo(0,ClientHeight-1);
    Brush.Color := clWindow;
  end;
end;

end.
