unit CoreOverview;

(***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************)

interface

uses
  Mars,
  Graphics,
  Classes,
  Controls,
  ExtCtrls;

type
  TCoreOverview = class(TCustomControl)
    protected
      FSelStart,
      FSelEnd: record // we have two values of each because one is a wrap value
        a, b: integer; // user selection
      end;
      FEcho: boolean;
      FCursor: TOperand;
      FCursorColour: TColor;
      FCursorState: boolean;
      FCursorBlink: TTimer;
      FCols,
      FCoreWidth,
      FCoreHeight: integer;
      FMars: TMars;
      function OnCoreChange(const ACore: TInstruction): boolean;
      procedure OnStep(const AThread: Mars.TThread);
      procedure HandleResize(Sender: TObject);
      procedure OnCursorBlink(Sender: TObject);
      procedure PaintCell(AIndex: TOperand);
      procedure PaintAll;
      procedure PaintCursor;
      function GetMouseUp: TMouseEvent;
      procedure SetMouseUp(AMouseUp: TMouseEvent);
    public
      constructor Create(AMars: TMars; AOwner: TComponent); reintroduce;
      destructor Destroy; override;
      procedure SetSelection(AStart, AEnd: integer);
      procedure SetCursor(ACursor: TOperand);
      function PointToIndex(X,Y: integer): TOperand; // returns -1 if no hit
      procedure Paint; override;
      property OnMouseUp: TMouseEvent read GetMouseUp write SetMouseUp;
    end;

implementation

uses
  GlobalConfig,
  Windows;

(******* TCoreOverview implementation ********************************)

constructor TCoreOverview.Create(AMars: TMars; AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque];
  DoubleBuffered := GlobalConfig.DoubleBuffered;
  FMars := AMars;
  SetSelection(-1,-1); // nothing
  SetCursor(-1);
  FCursorState := True;
  FCursorBlink := TTimer.Create(Self);
  with FCursorBlink do begin
    Interval := 500; // half a second
    OnTimer := OnCursorBlink;
    Enabled := True;
  end;
  // listeners
  OnResize := HandleResize;
  FMars.AddCoreListener(OnCoreChange);
  FMars.AddStepListener(OnStep);
end;

destructor TCoreOverview.Destroy;
begin
  // listeners
  FMars.RemoveCoreListener(OnCoreChange);
  inherited;
end;

function TCoreOverview.GetMouseUp: TMouseEvent;
begin
  Result := inherited OnMouseUp;
end;

procedure TCoreOverview.SetMouseUp(AMouseUp: TMouseEvent);
begin
  inherited OnMouseUp := AMouseUp;
end;

procedure TCoreOverview.SetSelection(AStart, AEnd: integer);
begin
  FSelStart.a := AStart;
  FSelEnd.a := AEnd;
  if (FSelStart.a > FMars.Fold(FSelEnd.a)) then begin // wrapped selection past end?
    FSelStart.b := 0;
    FSelEnd.b := FMars.Fold(FSelEnd.a);
    FSelEnd.a := FMars.CORESIZE;
  end else if ((FSelStart.a < 0) and (FSelEnd.a >= 0)) then begin // wrapped past beginning?
    FSelStart.b := FMars.Fold(FSelStart.a);
    FSelStart.a := 0;
    FSelEnd.b := FMars.CORESIZE;
  end else begin // can null the second selection
    FSelStart.b := -1;
    FSelEnd.b := -1;
  end;
  Repaint;
end;

procedure TCoreOverview.SetCursor(ACursor: TOperand);
begin
  FCursor := ACursor;
end;

procedure TCoreOverview.PaintCursor;
var
  Cols,
  X, Y: integer;
  r: TRect;
begin
  if (FCoreWidth = 0) then // no point if we will get a div-zero exception
    Exit;
  if (FCursor <> -1) then begin
    Cols := Width div FCoreWidth;
    X := (FCursor mod Cols) * FCoreWidth;
    Y := (FCursor div Cols) * FCoreHeight;
    r := Rect(X,Y,X+FCoreWidth,Y+FCoreHeight);
    InvalidateRect(Handle,@r,False);
  end;
end;

procedure TCoreOverview.OnStep(const AThread: Mars.TThread);
function Brighten(ACol: TColor): TColor;
const
  BRIGHTER = 30;
var
  rgb: packed array[0..3] of byte absolute Result;
begin
  Result := ACol;
  Inc(rgb[1],BRIGHTER);
  Inc(rgb[2],BRIGHTER);
  Inc(rgb[3],BRIGHTER);
end;
begin
  // disabled echo?
  if (FMars.State = QUICK_RUN) then
    Exit;
  // repaint over old cursor?
  PaintCursor;
  // set the cursor
  if Assigned(AThread) then begin
    FCursorColour := Brighten(AThread.Process.Warrior.ID);
    SetCursor(AThread.PC);
  end else begin
    FCursorColour := clDkGray;
    SetCursor(-1);
  end;
  // Repaint over new cursor?
  PaintCursor;
end;

procedure TCoreOverview.OnCursorBlink(Sender: TObject);
begin
  if Visible then begin
    FCursorState := not FCursorState;
    PaintCell(FCursor);
  end;
end;

procedure TCoreOverview.HandleResize(Sender: TObject);
const
  MINCOREWIDTH = 5;
  MINCOREHEIGHT = 5;
var
  i,j: integer;
begin
  FCoreWidth := MINCOREWIDTH;
  FCoreHeight := MINCOREHEIGHT;
  j := 0;
  i := (Width div FCoreWidth) * (Height div FCoreHeight);
  while (i > FMars.CORESIZE) do begin
    Inc(j);
    if ((j mod 2) = 1) then
      Inc(FCoreWidth)
    else
      Inc(FCoreHeight);
    i := (Width div FCoreWidth) * (Height div FCoreHeight);
  end;
  // too big? scale back a bit if we can
  if (j > 0) and (i < FMARS.CORESIZE) then
    if ((j mod 2) = 1) then
      Dec(FCoreWidth)
    else
      Dec(FCoreHeight);
  // work out how many columns we have
  FCols := Width div FCoreWidth;
end;

procedure TCoreOverview.PaintCell(AIndex: TOperand);
var
  X, Y: integer;
  Instruction: TInstruction;
  Cell: TRect;
begin
  X := ((AIndex mod FCols) * FCoreWidth);
  Y := ((AIndex div FCols) * FCoreHeight);
  with Canvas do begin
    // get the cell
    Instruction := FMars.Core[AIndex];
    // work out the bounds
    Cell := Rect(X,Y,X+FCoreWidth,Y+FCoreHeight);
    // background
    Brush.Color := clDkGray;
    FrameRect(Cell);
    // prepare for colouring
    Inc(Cell.Left);
    Inc(Cell.Top);
    Dec(Cell.Right);
    Dec(Cell.Bottom);
    if not Assigned(Instruction.Owner) then begin // not owned?
      if Instruction.ContainsCode then // instruction?
        Brush.Color := clLtGray
      else // data then
        Brush.Color := clCream;
      FillRect(Cell);
    end else begin
      Brush.Color := TColor(Instruction.Owner.ID);
      if Instruction.ContainsCode then // instruction?
        FillRect(Cell)
      else begin // data then
        // coloured border
        FrameRect(Cell);
        // hollow centre
        Inc(Cell.Left);
        Inc(Cell.Top);
        Dec(Cell.Right);
        Dec(Cell.Bottom);
        Brush.Color := clLtGray;
        FillRect(Cell);
      end;
    end;
    // cursor blink?
    if (AIndex = FCursor) and FCursorState then begin
      Brush.Color := FCursorColour;
      FillRect(Cell);
    end;
  end;
end;

procedure TCoreOverview.PaintAll;
var
  i, X, Y, Col: integer;
begin
  X := 0; Col := 0;
  Y := 0;
  with Canvas do
    for i := 0 to Pred(FMars.CORESIZE) do begin
      // visible?
      if (X <= ClipRect.Right) and ((X+FCoreWidth) >= ClipRect.Left) and
        (Y <= ClipRect.Bottom) and ((Y+FCoreHeight) >= ClipRect.Top) then
        PaintCell(i);
      // move to next cell
      Inc(Col);
      if (Col >= FCols) then begin
        X := 0; Col := 0;
        Inc(Y,FCoreHeight);
      end else
        Inc(X,FCoreWidth);
    end;
end;

procedure TCoreOverview.Paint;
begin
  PaintAll;
end;

function TCoreOverview.PointToIndex(X,Y: integer): TOperand; // returns -1 if no hit
var
  Cols: integer;
begin
  Cols := Width div FCoreWidth;
  X := X div FCoreWidth;
  Y := Y div FCoreHeight;
  Result := Y * Cols + X;
  if (Result < 0) or (Result > FMars.CORESIZE) then
    Result := -1;
end;

function TCoreOverview.OnCoreChange(const ACore: TInstruction): boolean;
var
  Cols,
  X, Y: integer;
  r: TRect;
begin
  Result := False;
  if (FCoreWidth = 0) then // no point if we will get a div-zero exception
    Exit;
  if (ACore.Index <> -1) then
    PaintCell(ACore.Index);
end;

end.
