unit WarriorsList;

(***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************)

interface

uses
  Mars,
  Classes,
  Graphics,
  Types,
  StdCtrls,
  Controls;

type
  TWarriorsList = class(TCustomListBox)
    private
      FMars: TMars;
      FActiveProcess: TProcess;
      FEcho: boolean;
      procedure DrawWarriorItem(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState);
      procedure DrawCursor; overload;
      procedure DrawCursor(AIndex: integer); overload;
      procedure HandleStep(const AThread: Mars.TThread);
      function HandleWarrior(const AProcess: TProcess): boolean;
      procedure HandleStateChange(const AState: TRunState);
    public
      constructor Create(AMars: TMars; AParent: TWinControl; AOwner: TComponent); reintroduce;
      destructor Destroy; override;
    end;

implementation

uses
  Windows;

const
  GUTTER = 15;

constructor TWarriorsList.Create(AMars: TMars; AParent: TWinControl; AOwner: TComponent);
begin
  // my init
  inherited Create(AOwner);
  Parent := AParent;
  // mars stuff
  FMars := AMars;
  FMars.AddWarriorListener(HandleWarrior);
  FMars.AddStepListener(HandleStep);
  FMars.AddStateChangeListener(HandleStateChange);
  FActiveProcess := nil;
  FEcho := True;
  // my gui stuff
  MultiSelect := False;
  Style := lbOwnerDrawFixed;
  OnDrawItem := DrawWarriorItem;
end;

destructor TWarriorsList.Destroy;
begin
  FMars.RemoveWarriorListener(HandleWarrior);
  inherited;
end;

procedure TWarriorsList.HandleStateChange(const AState: TRunState);
begin
  FEcho := (AState in [RUN_STEP,STOP]);
  if FEcho then
    Repaint;
end;

procedure TWarriorsList.HandleStep(const AThread: Mars.TThread);
begin
  if FEcho then begin
    if Assigned(AThread) then
      FActiveProcess := AThread.Process;
    // else determine winner!
    DrawCursor;
  end;
end;

procedure TWarriorsList.DrawWarriorItem(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState);
var
  s: string;
  Proc: TProcess;
  Thread: Mars.TThread;
  War: TWarrior;
begin
  s := Items[Index];
  Proc := TProcess(Items.Objects[Index]);
  Thread := Proc.ActiveThread;
  War := Proc.Warrior;
  with Canvas do begin
    Brush.Color := clWindow;
    Font.Color := War.ID;
    if Proc.Alive then
      Font.Style := []
    else
      Font.Style := [fsItalic];
    // fill background
    FillRect(Rect);
    // text
    TextRect(Rect,GUTTER,Rect.Top,s);
    // selected?
    if (odSelected in State) then begin
      Brush.Color := clActiveCaption;
      Inc(Rect.Left);
      Inc(Rect.Top);
      Dec(Rect.Right);
      Dec(Rect.Bottom);
      FrameRect(Rect);
    end;
  end;
  // draw cursor
  DrawCursor(Index);
end;

procedure TWarriorsList.DrawCursor;
var
  i: integer;
begin
  for i := 0 to Succ(Height div ItemHeight) do
    DrawCursor(i+TopIndex);
end;

procedure TWarriorsList.DrawCursor(AIndex: integer);
const
  MARGIN = GUTTER div 5;
var
  Y: integer;
  Proc: TProcess;
begin
  if (AIndex in [0..Pred(Count)]) then
    Proc := TProcess(Items.Objects[AIndex])
  else
    Proc := nil;
  Y := ((AIndex-TopIndex)*ItemHeight);
  with Canvas do begin
    Brush.Color := clWindow;
    // clear gutter
    FillRect(Rect(0,Y,Pred(GUTTER),Y+ItemHeight));
    // is it active? draw cursor then
    if (Proc = FActiveProcess) then begin
      Brush.Color := TColor(FActiveProcess.Warrior.ID);
      Polygon([Point(MARGIN,Succ(Y)),Point(GUTTER-MARGIN,Y+(ItemHeight div 2)),
        Point(MARGIN,Pred(Y+ItemHeight))]);
      Brush.Color := clWindow;
    end;
  end;
end;

function TWarriorsList.HandleWarrior(const AProcess: TProcess): boolean;
var
  i: integer;
  Found: boolean;
begin
  // find warrior
  Found := False;
  for i := 0 to Pred(Count) do begin
    if (items.Objects[i] = AProcess) then begin
      Found := True;
      Break;
    end;
  end;
  if not Found then begin // new warrior added?
    AddItem(AProcess.Warrior.Name,AProcess);
  end;
  Repaint;
  Result := False;
end;

end.
