program validate;

(***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************)

{$APPTYPE CONSOLE}

uses
  Graphics,
  SysUtils,
  Breakpoint in '..\engine-common\Breakpoint.pas',
  Mars in '..\engine-common\Mars.pas',
  Parser in '..\engine-common\Parser.pas',
  Assembler in '..\engine-common\Assembler.pas';

const
  MAX_WARRIORS = 1; // 0-based! can increase this later, when we have more colours etc
  WarriorColours: array[0..MAX_WARRIORS] of TColor = (clBlue,clRed);

var
  // global constants (can be set by the command line)
  CoreSize: integer = 8000;
  MaxCycles: integer = 80000;
  MaxThreads: integer = 1000;
  // the warriors for this simulation, and their locations on the command line
  StartPos: array of integer;
  Warriors: array of TWarrior;
  // the global variables that represent this simulation
  MyMars: TMars;
  MyBreakpoints: TBreakpointManager;

procedure ParseCommandLine;
var
  Cmd: integer; // current command index
  S: string; // the actual command
function HasMoreParams: boolean; // returns true if there are more parameters
begin Result := (Cmd <= ParamCount); end;
function NextParam(AField: string): string; // displays an error message incorporating AField if there isn't another parameter
begin
  if not HasMoreParams then begin
    WriteLn('Error in commandline parameters: ',AField,' expected.');
    Halt(1);
  end;
  S := Trim(ParamStr(Cmd));
  Inc(Cmd);
  Result := S;
end;
procedure AddWarrior(AFilename: string; AStartPos: integer);
var
  i: integer;
begin
  i := Length(Warriors);
  if (i > MAX_WARRIORS) then begin
    WriteLn('Error in commandline parameters: too many warriors specified.');
    Halt(3);
  end;
  SetLength(Warriors,Succ(i));
  SetLength(StartPos,Succ(i));
  Warriors[i] := ParseRC(WarriorColours[i],AFilename);
  StartPos[i] := AStartPos;
end;
var
  Filename: string;
  i: integer;
begin
  if (ParamCount = 0) then begin
    WriteLn('Usage: validate [options] Warrior1.rc warrior1-startpos {Warrior2.rc warrior2-startpos}');
    WriteLn(#9,'options:');
    WriteLn(#9,#9,'-s coresize',#9,'(default = ',CoreSize,')');
    WriteLn(#9,#9,'-c cycles',#9,'(maximum cycles per warrior, default = ',MaxCycles,')');
    WriteLn(#9,#9,'-p procs',#9,'(maximum processes per warrior, default = ',MaxThreads,')');
    WriteLn(#9,'examples:');
    WriteLn(#9,#9,'validate mywarrior.rc 2400 npaper2.rc 6023');
    WriteLn(#9,#9,#9,'(would fight mywarrior against npaper2, mywarrior starting at 2400');
    WriteLn(#9,#9,#9,'and npaper2 starting at 6023)');
    Halt(0);
  end;
  // prepare
  Cmd := 1;
  SetLength(Warriors,0);
  SetLength(StartPos,0);
  // parse the commands
  while HasMoreParams do begin
    NextParam('warrior');
    if (S[1] = '-') then begin // a switch?
      if (S = '-s') then begin
        NextParam('size of core');
        CoreSize := StrToInt(S);
      end else if (S = '-c') then begin
        NextParam('maximum number of cycles');
        MaxCycles := StrToInt(S);
      end else if (S = '-p') then begin // see that 'processes' on the commandline => 'threads' in my MARS
        NextParam('maximum number of processes per warrior');
        MaxThreads := StrToInt(S);
      end else begin // bad switch!
        WriteLn('Error in commandline parameters: switch "',S,'" not understood (switches are case-sensitive!).');
        Halt(4);
      end;
    end else begin // a warrior!
        Filename := S;
        if not FileExists(Filename) then begin
          WriteLn('Error in commandline paramters: warrior "',Filename,'" does not exist!');
          Halt(5);
        end;
        NextParam('warrior starting offset');
        i := StrToInt(S);
        AddWarrior(Filename,i);
    end;
  end;
  // check we loaded some warriors
  if (Length(Warriors) = 0) then begin
    WriteLn('Error in commandline parameters: no warriors specified.');
    Halt(2);
  end;
  // done
end;

var
  i: integer;
begin
  try
    ParseCommandLine;
    // create mars
    MyMars := TMars.Create(CORESIZE,MAXCYCLES,MAXTHREADS);
    // add the warriors specified on the commandline
    for i := Low(Warriors) to High(Warriors) do
      MyMars.AddWarrior(Warriors[i],StartPos[i]);
    // run it
    MyMars.Run;
  except
    on E: Exception do begin
      WriteLn(E.Message);
      Halt(100);
    end;
  end;
end.
